@@ -1,9 +1,29 @@
-/* od.c -- Open Directory routines
+/*
+ * Copyright (c) 2004 Apple Computer, Inc. All rights reserved.
  *
- * Copyright (c) 2003 by Apple Computer, Inc., all rights reserved.
+ * @APPLE_LICENSE_HEADER_START@
  *
+ * This file contains Original Code and/or Modifications of Original Code
+ * as defined in and that are subject to the Apple Public Source License
+ * Version 2.0 (the 'License'). You may not use this file except in
+ * compliance with the License. Please obtain a copy of the License at
+ * http://www.opensource.apple.com/apsl/ and read it before using this
+ * file.
+ *
+ * The Original Code and all software distributed under the License are
+ * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
+ * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
+ * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
+ * Please see the License for the specific language governing rights and
+ * limitations under the License.
+ *
+ * @APPLE_LICENSE_HEADER_END@
  */
 
+/*
+ * od.c -- Open Directory routines
+ */
 
 
 #include <string.h>
@@ -62,6 +82,7 @@ int od_do_auth( const char *inMethod,
 	int				respLen	= 0;
 	char		   *ptr		= NULL;
 	char			user	[ MAX_USER_BUF_SIZE ];
+	char			prinUser[ MAX_USER_NAME_SIZE + 1 ];
 	char			passwd	[ MAX_USER_BUF_SIZE ];
 	char			chal	[ MAX_CHAL_BUF_SIZE ];
 	char			resp	[ MAX_IO_BUF_SIZE ];
@@ -69,12 +90,16 @@ int od_do_auth( const char *inMethod,
 	char			hostname[ MAXHOSTNAMELEN + 1 ];
 	gss_buffer_desc	in_token;
 	gss_buffer_desc	out_token;
+	gss_buffer_desc	disp_name;
+	gss_OID			mech_type;
 	OM_uint32		minStatus	= 0;
 	OM_uint32		majStatus	= 0;
 	gss_ctx_id_t	context		= GSS_C_NO_CONTEXT;
 	OM_uint32		ret_flags	= 0;
 	gss_name_t		clientName;
 	unsigned long	maxsize		= htonl( 8192 );
+	krb5_context	krb5Context;
+	krb5_principal	krb5Principal;
 
 	if ( strcasecmp( inMethod, "cram-md5" ) == 0 )
 	{
@@ -316,193 +341,235 @@ int od_do_auth( const char *inMethod,
 	else if ( strcasecmp( inMethod, "GSSAPI" ) == 0 )
 	{
 		r = gss_Init( inProtocol );
-		if ( r == GSS_S_COMPLETE )
+		if ( r != GSS_S_COMPLETE )
 		{
-			r = ODA_AUTH_FAILED;
-			prot_printf( inStreamOut, "+ \r\n" );
-			if ( prot_fgets( buff, MAX_IO_BUF_SIZE, inStreamIn ) )
-			{
-				/* trim CRLF */
-				ptr = buff + strlen( buff ) - 1;
-				if ( (ptr >= buff) && (*ptr == '\n') )
-				{
-					*ptr-- = '\0';
-				}
-				if ( (ptr >= buff) && (*ptr == '\r') )
-				{
-					*ptr-- = '\0';
-				}
+			gss_release_cred( &minStatus, &stCredentials );
 
-				/* check if client cancelled */
-				if ( buff[ 0 ] == '*' )
-				{
-					return( ODA_AUTH_CANCEL );
-				}
+			return ( ODA_PROTOCOL_ERROR );
+		}
+
+		/* set default auth failure */
+		r = ODA_AUTH_FAILED;
+		prot_printf( inStreamOut, "+ \r\n" );
+		if ( prot_fgets( buff, MAX_IO_BUF_SIZE, inStreamIn ) )
+		{
+			/* trim CRLF */
+			ptr = buff + strlen( buff ) - 1;
+			if ( (ptr >= buff) && (*ptr == '\n') )
+			{
+				*ptr-- = '\0';
+			}
+			if ( (ptr >= buff) && (*ptr == '\r') )
+			{
+				*ptr-- = '\0';
 			}
 
-			/* clear response buffer */
-			memset( resp, 0, MAX_IO_BUF_SIZE );
+			/* check if client cancelled */
+			if ( buff[ 0 ] == '*' )
+			{
+				gss_release_cred( &minStatus, &stCredentials );
+				return( ODA_AUTH_CANCEL );
+			}
+		}
 
-			len = strlen( buff );
-			aodDecodeBase64( buff, len, resp, MAX_IO_BUF_SIZE, &respLen );
+		/* clear response buffer */
+		memset( resp, 0, MAX_IO_BUF_SIZE );
 
-			in_token.value  = resp;
-			in_token.length = respLen;
+		len = strlen( buff );
+		aodDecodeBase64( buff, len, resp, MAX_IO_BUF_SIZE, &respLen );
 
-			do {
-				// negotiate authentication
-				majStatus = gss_accept_sec_context(	&minStatus,
-													&context,
-													stCredentials,
-													&in_token,
-													GSS_C_NO_CHANNEL_BINDINGS,
-													&clientName,
-													NULL, /* &mechTypes */
-													&out_token,
-													&ret_flags,
-													NULL,	/* ignore time?*/
-													NULL );
+		in_token.value  = resp;
+		in_token.length = respLen;
 
-				switch ( majStatus )
+		do {
+			// negotiate authentication
+			majStatus = gss_accept_sec_context(	&minStatus,
+												&context,
+												stCredentials,
+												&in_token,
+												GSS_C_NO_CHANNEL_BINDINGS,
+												&clientName,
+												&mech_type,
+												&out_token,
+												&ret_flags,
+												NULL, NULL );
+
+			switch ( majStatus )
+			{
+				case GSS_S_COMPLETE:		// successful
+				case GSS_S_CONTINUE_NEEDED:
 				{
-					case GSS_S_COMPLETE:		// successful
-					case GSS_S_CONTINUE_NEEDED:
+					if ( out_token.value )
 					{
-						if ( out_token.value )
+						// Encode the challenge and send it
+						aodEncodeBase64( (char *)out_token.value, out_token.length, buff, MAX_IO_BUF_SIZE );
+
+						prot_printf( inStreamOut, "+ %s\r\n", buff );
+						if ( prot_fgets( buff, MAX_IO_BUF_SIZE, inStreamIn ) )
 						{
-							// Encode the challenge and send it
-							aodEncodeBase64( (char *)out_token.value, out_token.length, buff, MAX_IO_BUF_SIZE );
-
-							prot_printf( inStreamOut, "+ %s\r\n", buff );
-							if ( prot_fgets( buff, MAX_IO_BUF_SIZE, inStreamIn ) )
+							/* trim CRLF */
+							ptr = buff + strlen( buff ) - 1;
+							if ( (ptr >= buff) && (*ptr == '\n') )
 							{
-								/* trim CRLF */
-								ptr = buff + strlen( buff ) - 1;
-								if ( (ptr >= buff) && (*ptr == '\n') )
-								{
-									*ptr-- = '\0';
-								}
-								if ( (ptr >= buff) && (*ptr == '\r') )
-								{
-									*ptr-- = '\0';
-								}
-
-								/* check if client cancelled */
-								if ( buff[ 0 ] == '*' )
-								{
-									return( ODA_AUTH_CANCEL );
-								}
+								*ptr-- = '\0';
+							}
+							if ( (ptr >= buff) && (*ptr == '\r') )
+							{
+								*ptr-- = '\0';
 							}
 
-							memset( resp, 0, MAX_IO_BUF_SIZE );
-
-							// Decode the response
-							len = strlen( buff );
-							aodDecodeBase64( buff, len, resp, MAX_IO_BUF_SIZE, &respLen );
-
-							in_token.value  = resp;
-							in_token.length = respLen;
-
-							gss_release_buffer( &minStatus, &out_token );
+							/* check if client cancelled */
+							if ( buff[ 0 ] == '*' )
+							{
+								gss_release_cred( &minStatus, &stCredentials );
+								return( ODA_AUTH_CANCEL );
+							}
 						}
-						break;
+
+						memset( resp, 0, MAX_IO_BUF_SIZE );
+
+						// Decode the response
+						len = strlen( buff );
+						aodDecodeBase64( buff, len, resp, MAX_IO_BUF_SIZE, &respLen );
+
+						in_token.value  = resp;
+						in_token.length = respLen;
+
+						gss_release_buffer( &minStatus, &out_token );
 					}
-
-					default:
-						log_errors( "gss_accept_sec_context", majStatus, minStatus );
-						break;
+					break;
 				}
-			} while ( in_token.value && in_token.length && (majStatus == GSS_S_CONTINUE_NEEDED) );
 
+				default:
+					log_errors( "gss_accept_sec_context", majStatus, minStatus );
+					break;
+			}
+		} while ( in_token.value && in_token.length && (majStatus == GSS_S_CONTINUE_NEEDED) );
+
+		if ( majStatus == GSS_S_COMPLETE )
+		{
+			gss_buffer_desc		inToken;
+			gss_buffer_desc		outToken;
+
+			memcpy( passwd, (void *)&maxsize, 4 );
+			inToken.value	= passwd;
+			inToken.length	= 4;
+
+			passwd[ 0 ] = 1;
+
+			majStatus = gss_wrap( &minStatus, context, 0, GSS_C_QOP_DEFAULT, &inToken, NULL, &outToken );
 			if ( majStatus == GSS_S_COMPLETE )
 			{
-				gss_buffer_desc		inToken;
-				gss_buffer_desc		outToken;
+				// Encode the challenge and send it
+				aodEncodeBase64( (char *)outToken.value, outToken.length, buff, MAX_IO_BUF_SIZE );
 
-				memcpy( passwd, (void *)&maxsize, 4 );
-				inToken.value	= passwd;
-				inToken.length	= 4;
-
-				passwd[ 0 ] = 1;
-
-				majStatus = gss_wrap( &minStatus, context, 0, GSS_C_QOP_DEFAULT, &inToken, NULL, &outToken );
-				if ( majStatus == GSS_S_COMPLETE )
+				prot_printf( inStreamOut, "+ %s\r\n", buff );
+				memset( buff, 0, MAX_IO_BUF_SIZE );
+				if ( prot_fgets( buff, MAX_IO_BUF_SIZE, inStreamIn ) )
 				{
-					// Encode the challenge and send it
-					aodEncodeBase64( (char *)outToken.value, outToken.length, buff, MAX_IO_BUF_SIZE );
-
-					prot_printf( inStreamOut, "+ %s\r\n", buff );
-					memset( buff, 0, MAX_IO_BUF_SIZE );
-					if ( prot_fgets( buff, MAX_IO_BUF_SIZE, inStreamIn ) )
+					/* trim CRLF */
+					ptr = buff + strlen( buff ) - 1;
+					if ( (ptr >= buff) && (*ptr == '\n') )
 					{
-						/* trim CRLF */
-						ptr = buff + strlen( buff ) - 1;
-						if ( (ptr >= buff) && (*ptr == '\n') )
-						{
-							*ptr-- = '\0';
-						}
-						if ( (ptr >= buff) && (*ptr == '\r') )
-						{
-							*ptr-- = '\0';
-						}
-
-						/* check if client cancelled */
-						if ( buff[ 0 ] == '*' )
-						{
-							return( ODA_AUTH_CANCEL );
-						}
+						*ptr-- = '\0';
+					}
+					if ( (ptr >= buff) && (*ptr == '\r') )
+					{
+						*ptr-- = '\0';
 					}
 
-
-					// Decode the response
-					respLen = 0;
-					memset( resp, 0, MAX_IO_BUF_SIZE );
-					len = strlen( buff );
-
-					aodDecodeBase64( buff, len, resp, MAX_IO_BUF_SIZE, &respLen );
-
-					inToken.value  = resp;
-					inToken.length = respLen;
-
-					gss_release_buffer( &minStatus, &outToken );
-
-					majStatus = gss_unwrap( &minStatus, context, &inToken, &outToken, NULL, NULL );
-					if ( majStatus == GSS_S_COMPLETE )
+					/* check if client cancelled */
+					if ( buff[ 0 ] == '*' )
 					{
-						if ( (outToken.value != NULL)		&&
-							(outToken.length > 4)			&&
-							(outToken.length < MAX_USER_BUF_SIZE) )
+						gss_release_cred( &minStatus, &stCredentials );
+						return( ODA_AUTH_CANCEL );
+					}
+				}
+
+				// Decode the response
+				respLen = 0;
+				memset( resp, 0, MAX_IO_BUF_SIZE );
+				len = strlen( buff );
+
+				aodDecodeBase64( buff, len, resp, MAX_IO_BUF_SIZE, &respLen );
+
+				inToken.value  = resp;
+				inToken.length = respLen;
+
+				gss_release_buffer( &minStatus, &outToken );
+
+				majStatus = gss_unwrap( &minStatus, context, &inToken, &outToken, NULL, NULL );
+				if ( majStatus == GSS_S_COMPLETE )
+				{
+					if ( (outToken.value != NULL) &&
+						 (outToken.length > 4)	  &&
+						 (outToken.length < MAX_USER_BUF_SIZE) )
+					{
+						memcpy( user, outToken.value, outToken.length );
+						if ( user[0] & 1 )
 						{
-							memcpy( user, outToken.value, outToken.length );
-							if ( user[0] & 1 )
+							user[ outToken.length ] = '\0';
+							if ( gss_display_name( &minStatus, clientName, &disp_name, &mech_type ) == GSS_S_COMPLETE  )
 							{
-								user[ outToken.length ] = '\0';
-								*inOutCannonUser = auth_canonifyid( user + 4, 0 );
-								r = kSGSSSuccess;
+								if ( !krb5_init_context( &krb5Context ) )
+								{
+									if ( !krb5_parse_name( krb5Context, disp_name.value, &krb5Principal ) )
+									{
+										if ( !krb5_aname_to_localname( krb5Context, krb5Principal, MAXHOSTNAMELEN-1, prinUser ) )
+										{
+											const char *p = user+4;
+											if ( !strcmp( p, prinUser ) )
+											{
+												*inOutCannonUser = auth_canonifyid( user + 4, 0 );
+												r = kSGSSSuccess;
+											}
+											else
+											{
+												syslog( LOG_NOTICE, "GSSAPI Error: principal user and ticket mismatch: %s != %s.", p, prinUser );
+											}
+										}
+										else
+										{
+											syslog( LOG_NOTICE, "GSSAPI Error: krb5_aname_to_localname failed." );
+										}
+										krb5_free_principal( krb5Context, krb5Principal );
+									}
+									else
+									{
+										syslog( LOG_NOTICE, "GSSAPI Error: krb5_parse_name failed." );
+									}
+									krb5_free_context( krb5Context );	/* finished with context */
+								}
+								else
+								{
+									syslog( LOG_NOTICE, "GSSAPI Error: krb5_init_context failed." );
+								}
+							}
+							else
+							{
+								log_errors( "gss_display_name", majStatus, minStatus );
 							}
 						}
 					}
-					else
-					{
-						log_errors( "gss_unwrap", majStatus, minStatus );
-					}
-					gss_release_buffer( &minStatus, &outToken );
 				}
 				else
 				{
-					log_errors( "gss_wrap", majStatus, minStatus );
+					log_errors( "gss_unwrap", majStatus, minStatus );
 				}
+				gss_release_buffer( &minStatus, &outToken );
 			}
 			else
 			{
-				log_errors( "gss_accept_sec_context", majStatus, minStatus );
+				log_errors( "gss_wrap", majStatus, minStatus );
 			}
 		}
+		else
+		{
+			log_errors( "gss_accept_sec_context", majStatus, minStatus );
+		}
 	}
-	else
-	{
-		r = ODA_PROTOCOL_ERROR;
-	}
+
+	gss_release_cred( &minStatus, &stCredentials );
 
     return ( r );
 
@@ -542,7 +609,7 @@ int gss_Init ( const char *inProtocol )
 	{
 		majStatus = gss_acquire_cred( &minStatus, 
 										principalName, 
-										GSS_C_INDEFINITE, 
+										0, /* GSS_C_INDEFINITE */
 										GSS_C_NO_OID_SET,
 										GSS_C_ACCEPT,
 									   &stCredentials,
@@ -728,3 +795,4 @@ void log_errors ( char *inName, OM_uint32 inMajor, OM_uint32 inMinor )
 	} while ( msg_context != 0 );
 
 } // LogErrors
+
